#include "cache.h"
#include <stdlib.h>

static double urandom(void){
  double r = random();
  return r / 2147483648.0; /* 2^31 */
}

int print_list(Cache *cache){
  Trace_list *temp;

  temp = cache->list;
  fprintf(stdout,"list :\n");
  while(temp){
    fprintf(stdout, "\trecord: %d, time %f: , size: %d, prefetch: %d\n",
            temp->record.blkno, temp->record.start,temp->record.bcount ,
            temp->record.prefetch );
    temp = temp->next;
  }
  fflush(stdout);
  return 0;
}

void push_queue(Trace_list *q, Request *r){
  Trace_list *temp, *temp_next;

  temp = temp_next = q;
  if(!temp){
    q = malloc(sizeof(Trace_list));
    q->record.start = r->start;
    q->record.flags = r->flags;
    q->record.app = r->app;
    q->record.devno = r->devno;
    q->record.blkno = r->blkno;
    q->record.bcount = r->bcount;
    q->record.clientno = r->clientno;
    q->next = NULL;
  }
  else{
    while(temp_next && (temp_next->record.start +0.0001) < r->start){
      temp = temp_next;
      temp_next = temp->next;
    }
    temp->next = malloc(sizeof(Trace_list)); 
    temp->next->record.start = r->start;
    temp->next->record.flags = r->flags;
    temp->next->record.app = r->app;
    temp->next->record.devno = r->devno;
    temp->next->record.blkno = r->blkno;
    temp->next->record.bcount = r->bcount;
    temp->next->record.clientno = r->clientno;
    temp->next->next = temp_next;
  }
}

void pop_queue(Trace_list *q, Request *r){
  Trace_list *temp, *temp_next;

  r = &q->record;
  q = q->next;
}

int init_client_cache(Cache *cache, int clientno, int size, FILE *fd, struct disksim *_disksim){
  cache->clientno = clientno;
  cache->free_size = size;
  cache->app_number = 0;
  cache->completed = 1;
  cache->fd = fd;
  cache->last_prefetched = 0;
  cache->list = NULL;
  cache->_disksim = _disksim;
  return 0;
}

int close_client_cache(Cache *cache){
  Trace_list *temp, *temp_next;

  temp = cache->list;
  if(temp){
    temp_next = temp->next;
    free(temp);
    temp = temp->next;
  }
  return 0;
}

int init_disk_cache(Cache *cache, int diskno, int size, FILE *fd, struct disksim *_disksim){
  cache->clientno = -1;
  cache->diskno = diskno;
  cache->free_size = size;
  cache->app_number = 0;
  cache->completed = 1;
  cache->fd = fd;
  cache->last_prefetched = 0;
  cache->list = NULL;
  cache->_disksim = _disksim;
  return 0;
}  

int close_disk_cache(Cache *cache){
  Trace_list *temp, *temp_next;
  
  temp = cache->list;
  if(temp){
    temp_next = temp->next;
    free(temp);
    temp = temp->next;
  }
  return 0;
}

/*cache a record */
int add_record(Cache *cache, Request record){
  Trace_list *temp, *temp_next;
  int ok;

  ok = cache_lookup(cache, record);
  if(ok == 1) return 0;
  //print_list(cache);
  temp = temp_next = cache->list;
  if(!temp){
    cache->list = malloc(sizeof(Trace_list));
    cache->list->record.start = record.start;
    cache->list->record.flags = record.flags;
    cache->list->record.app = record.app;
    cache->list->record.devno = record.devno;
    cache->list->record.blkno = record.blkno;
    cache->list->record.bcount = record.bcount;
    cache->list->record.clientno = record.clientno;
    cache->list->record.prefetch = record.prefetch;
    cache->list->record.hit_count = record.hit_count;
    cache->list->next = NULL;
  }
  else{
    while(temp_next && (temp_next->record.start + 0.00001) < record.start){
      temp = temp_next;
      temp_next = temp->next;
    }
    if(temp_next == cache->list){
      cache->list =  malloc(sizeof(Trace_list));
      cache->list->record.start = record.start;
      cache->list->record.flags = record.flags;
      cache->list->record.app = record.app;
      cache->list->record.devno = record.devno;
      cache->list->record.blkno = record.blkno;
      cache->list->record.bcount = record.bcount;
      cache->list->record.clientno = record.clientno;
      cache->list->record.prefetch = record.prefetch;
      cache->list->record.hit_count = record.hit_count;
      cache->list->next = temp_next;
    }
    else{
      temp->next = malloc(sizeof(Trace_list)); 
      temp->next->record.start = record.start;
      temp->next->record.flags = record.flags;
      temp->next->record.app = record.app;
      temp->next->record.devno = record.devno;
      temp->next->record.blkno = record.blkno;
      temp->next->record.bcount = record.bcount;
      temp->next->record.clientno = record.clientno;
      cache->list->record.prefetch = record.prefetch;
      cache->list->record.hit_count = record.hit_count;
      temp->next->next = temp_next;
    }
  }
  cache->free_size = cache->free_size - record.bcount;

  return 0;
}

/*delete a record from cache*/
int delete_record(Cache * cache, Request record){
  Trace_list *temp, *temp_next;
  
  //fprintf(stderr, "delete record...%f, size of %d\n", record.start, record.bcount);
 
  temp = temp_next = cache->list;
  //print_list(cache);
  while(fabs(temp_next->record.start - record.start) > 0.00001){
    temp = temp_next;
    temp_next = temp->next;
  }
  if(temp_next == NULL){
    return -1;
  }
  else if(temp_next == cache->list){
    cache->list =  cache->list->next;
  }
  else{
    temp->next = temp_next->next;
  }

  cache->free_size = cache->free_size + record.bcount;
  free(temp_next);
  //print_list(cache);
 
  return 0;
}


int cache_lookup(Cache *cache, Request io_request){
  Trace_list *temp;
 
  temp = cache->list;
  while(temp){
    if((temp->record.devno == io_request.devno) &&(temp->record.blkno == io_request.blkno) && (temp->record.bcount >= io_request.bcount) ){
      return 1;
    }
    else temp = temp->next;
  }
  return 0;
}

int update_cache(Cache *cache, Request io_request){
  Trace_list *temp, *temp_list, *next_list;
  Request record;
  int ok;

  temp = cache->list;
  while(temp){
    if((temp->record.devno == io_request.devno) &&(temp->record.blkno == io_request.blkno) && (temp->record.bcount >= io_request.bcount) ){
      record.start = temp->record.start;
      record.flags = temp->record.flags;
      record.app = temp->record.app;
      record.devno = temp->record.devno;
      record.blkno = temp->record.blkno;
      record.bcount = temp->record.bcount;
      record.clientno = temp->record.clientno;
      record.hit_count = temp->record.hit_count;
      // print_list(cache);
      //fprintf(stderr, "delete while updating\n");
      ok = delete_record(cache, record);
      assert(ok != -1);
      //print_list(cache);
      record.start = io_request.start; 
      record.prefetch = io_request.prefetch;
 
      temp_list = cache->list;
      while(temp_list && cache->free_size < record.bcount){
	next_list = temp_list->next;
        //  fprintf(stderr, "cache free size %d, need %d\n", cache->free_size ,record.bcount);
        //fprintf(stderr, "delete when something is wrong\n");
	delete_record(cache, temp_list->record);
	temp_list = next_list;
      }

      // fprintf(stderr, "cache free size %d, need %d\n", cache->free_size ,record.bcount);
      assert(cache->free_size >= record.bcount);
      add_record(cache, record);
      //print_list(cache);
      break;
    }
    temp = temp->next;
  }
  return 0;
}  

int update_hit_rate(Cache *cache, Request io_request){
  Trace_list *temp, *temp_list, *next_list;
  Request record;
  int ok;
  
  temp = cache->list;
  while(temp){
      if((temp->record.devno == io_request.devno) &&(temp->record.blkno
                                                         ==
                                                     io_request.blkno)
         && (temp->record.bcount >= io_request.bcount) ){
          temp->record.start = io_request.start;
          temp->record.prefetch = 0;
          temp->record.hit_count ++;
          break;
      }
      temp = temp->next;
  }
  
  return 0;
}

                    
int set_fd(FILE *fd, Request * r){
  char line[201];
  Request new;
  
  while(fgets(line, 200, fd) != NULL) {
    if (sscanf(line, "%lf %d %d %d %x %d\n", &new.start, &new.devno, &new.blkno, &new.bcount, &new.flags, &new.clientno) != 6) {
      fprintf(stderr, "Wrong number of arguments for I/O trace event type\n");
      fprintf(stderr, "line: %s", line);
      exit(1);
    }
    else if(fabs(new.start - r->start) < 0.001){
      break;
    }
  }

  return 0;
}

/*return 1 if prefetch, 0 if cach is already full*/
int aggressive(Cache *cache, Request *record){
  char line[50];
  Request new;
  Trace_list *temp_list, *next_list;
  int ok;
  float r;

  r = urandom();
  if( r < cache->positive_error){
    disksim_io_request(cache->_disksim, record);
    return 1;
  }
  /* if(cache->last_prefetched < record->start){
      cache->last_prefetched = record->start; 
      }*/
  
  fseek(cache->fd, 0, SEEK_SET); /* this is stupid, but what else can i do? */
  while(fgets(line, 50, cache->fd) != NULL) {
    if (sscanf(line, "%lf %d %d %d %d %d\n", &new.start, &new.devno, &new.blkno, &new.bcount, &new.flags, &new.clientno) != 6) {
        fprintf(stderr, "Wrong number of arguments for I/O trace event type\n");
        fprintf(stderr, "line: %s", line);
      exit(1);
    }
    
    if(cache->last_prefetched == 0) break;
    if((cache->last_prefetched+ 0.00001 )< record->start  && fabs(record->start
                                                      -new.start) <
       0.00001) break;
    
    if((cache->last_prefetched + 0.00001 > record->start) && (cache->last_prefetched - new.start) < 0.00001){
        if(fgets(line, 200, cache->fd) != NULL){
            sscanf(line, "%lf %d %d %d %d %d\n", &new.start, &new.devno, &new.blkno, &new.bcount, &new.flags, &new.clientno);
        }
        else{
            return 0; /* end of trace*/
        }
        break;
    }
  }
  
  if(cache->clientno == -1){
    if(new.devno != cache->diskno) return 0;
  }
  else if(new.clientno != cache->clientno) return 0; /* only pretfetch at the proper client machine*/
  
  ok = cache_lookup(cache, new);
  if(ok == 1){
      if(cache->clientno == -1){/*disk side cache*/
          new.prefetch = 2;
      }
      else{
          new.prefetch = 1;
      }
      cache->last_prefetched = new.start;
      update_cache(cache, new);
      return 1;
  }
  if(ok != 1){
      temp_list = cache->list;
      while(temp_list && cache->free_size < new.bcount &&
            temp_list->record.prefetch == 0){
          /* fprintf(stderr, "freesize in the cache is %d, bcount is %d\n",
             cache->free_size , new.bcount);*/
          next_list = temp_list->next;
      if(temp_list->record.start < record->start && temp_list->record.prefetch
         == 0){
          delete_record(cache, temp_list->record);
      }
      temp_list = next_list;
    }
    
    if(cache->free_size >= new.bcount){
      if(cache->clientno == -1){/*disk side cache*/
	new.prefetch = 2;
      }
      else{
	new.prefetch = 1;
      }
      cache->last_prefetched = new.start;
      r = urandom();
      if(r > cache->negative_error){
          //fprintf(stderr, "now call back to nasd.c\n");
	disksim_io_request(cache->_disksim, &new);
      }
    }
    else{
      return 0;
    }
  }
  
  return 1;
}

int fixed_horizen(Cache *cache, Request *record, int total_hit){
  float benefit, lru_cost, r;
  float t_disk, t_hit, t_drive;
  int horizen, depth;
  char line[50];
  Request new;
  Trace_list *temp_list, *next_list;
  int ok, hit;
  int begin_count;

  r = urandom();
  if( r < cache->positive_error){
    disksim_io_request(cache->_disksim, record);
    return 1;
  }

  // if(cache->clientno != 3) return 0; // for debug reason, need to delete!!!!

  begin_count = 0;
   fseek(cache->fd, 0, SEEK_SET); /* this is stupid, but what else can i do? */
  while(fgets(line, 200, cache->fd) != NULL) {
    if (sscanf(line, "%lf %d %d %d %d %d\n", &new.start, &new.devno, &new.blkno, &new.bcount, &new.flags, &new.clientno) != 6) {
      fprintf(stderr, "Wrong number of arguments for I/O trace event type\n");
      fprintf(stderr, "line: %s", line);
      exit(1);
    }

    if(cache->last_prefetched == 0){
        depth = 0;
        break; 
    }

    if((cache->last_prefetched+ 0.0001) < record->start){
        if(fabs(record->start - new.start) < 0.00001){
            depth = 0;
            break;
        }
    }
    else{
        if(fabs(record->start - new.start) < 0.00001){
            depth = 0;
            begin_count = 1;
        }
        
        if((cache->last_prefetched - new.start) < 0.00001){
            if(fgets(line, 200, cache->fd) != NULL){
                sscanf(line, "%lf %d %d %d %d %d\n", &new.start, &new.devno, &new.blkno, &new.bcount, &new.flags, &new.clientno);
            }
            else{
                return 0; /* end of trace*/
            }
            break;
        }
        else if(begin_count == 1){
            depth ++;
        }
    }
  }

  if(cache->clientno == -1){
    if(new.devno != cache->diskno) return 0;
  }
  else if(new.clientno != cache->clientno) return 0; /* only pretfetch at the proper client machine*/ 
  //print_list(cache);

  /*The exact fixed-horizen algorithm is too complicated to implement and
    is not necessary in our case. In our case, the prefetch is always done
    according to the time order. So the benefit of a prefetch is alway less 
    than evicting a former prefetched record in the hinted cache. So the easy 
    way to do fixed-horizen is just to compare the benefit of bringing in a
    new record against the cost of LRU shrinkage*/
    
  t_disk = 0.002; /*temp assigment */
  t_hit = 0.0001;
  t_drive = 0.001;
 
  horizen = t_disk/t_hit;
  if(depth == 0){
    benefit = t_disk;
  }
  else if(depth < horizen){
    benefit = t_disk/(depth*(depth+1));
  }
  else{
    benefit = 0;
  }

  if(benefit <= 0) return 0;
  if(benefit > 0){
    ok = cache_lookup(cache, new);
    if(ok == 1){
        if(cache->clientno == -1){/*disk side cache*/
            new.prefetch = 2;
        }
        else{
            new.prefetch = 1;
      }
        cache->last_prefetched = new.start;
        //  fprintf(stderr, "start time %f\n",  new.start);
        update_cache(cache, new);
        return 1;
    }
    if(ok != 1){
      temp_list = cache->list;
      while(temp_list && cache->free_size < new.bcount){
          //fprintf(stderr, "freesize in the cache is %d, start is %f, bcount is %d\n", cache->free_size , new.start, new.bcount);
	/*if(temp_list->record.start < record->start){
	  delete_record(cache, temp_list->record);
	  }*/
        next_list = temp_list->next;
	if(temp_list->record.prefetch == 0){
            hit = temp_list->record.hit_count -1;
            hit = (hit < 0) ? 0 : hit;
            lru_cost = (t_disk + t_drive) *
                (hit/total_hit);
	  if(benefit > lru_cost){
              //  fprintf(stderr, "FOUND Better prefetch\n");
              delete_record(cache, temp_list->record);
	  }
        }
	
	temp_list = next_list;
      }
    
      if(cache->free_size >= new.bcount){
	if(cache->clientno == -1){
	  new.prefetch = 2;
	}
	else{
	  new.prefetch = 1;
	}
	cache->last_prefetched = new.start;
	r = urandom();
	if(r > cache->negative_error){
            // fprintf(stderr, "now call back to nasd.c\n");
	  disksim_io_request(cache->_disksim, &new);
	}
      }
      else{
	return 0;
      }
    }
  }
  // print_list(cache);
  return 1;
}

int fixed_fetch(Cache *cache, Request *record){
  

  return 0;
}

int no_prefetch(Cache *cache, Request *record){

  return 0;
}
