#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <math.h>
#include "nasd.h"
#include "disksim_rand48.h"
#include "cache.h"
#include "network.h"
//#include "disksim_global.h"
#include "disksim_iotrace.h"

#define	BLOCK	4096
#define	SECTOR	512
#define	BLOCK2SECTOR	(BLOCK/SECTOR)

typedef	struct	{
	int n;
	double sum;
	double sqr;
} Stat;


static SysTime now = 0;		/* current time */
static SysTime next_event = -1;	/* next event */
static int completed = 0;	/* last request was completed */
static Stat st;

Cache cache[10];

void
panic(const char *s)
{
	perror(s);
	exit(1);
}


void
add_statistics(Stat *s, double x)
{
	s->n++;
	s->sum += x;
	s->sqr += x*x;
}


void
print_statistics(FILE *fd, Stat *s, const char *title)
{
	double avg, std;

	avg = s->sum/s->n;
	std = sqrt((s->sqr - 2*avg*s->sum + s->n*avg*avg) / s->n);
	fprintf(fd, "%s: n=%d average=%f std. deviation=%f\n", title, s->n, avg, std);
}

/*parser. */
int parse_trace_file(FILE *fd, Request *req){
  ioreq_event *temp;
  
  temp = malloc(sizeof(ioreq_event));
  temp = iotrace_get_ioreq_event(fd, ASCII, temp);

  if(temp == NULL){
    return -1;
  };
  req->devno = temp->devno;
  req->blkno = temp->blkno;
  req->bcount = temp->bcount;
  req->flags = temp->flags;
  req->clientno = temp->clientno;
  req->start = temp->time;

  fprintf(stdout, "check check. blkno is %d\n", req->blkno);
  free(temp);
  return 0;
  
}

int process_next_trace(ioreq_event * item){
  return 0;
}
/*
 * Schedule next callback at time t.
 * Note that there is only *one* outstanding callback at any given time.
 * The callback is for the earliest event.
 */
void
syssim_schedule_callback(void (*f)(void *,double), SysTime t)
{
	next_event = t;
}


/*
 * de-scehdule a callback.
 */
void
syssim_deschedule_callback(void (*f)())
{
	next_event = -1;
}


void
syssim_report_completion(SysTime t, Request *r)
{
  float tmp;
  
  tmp = net_time();
  completed = 1;
  now = t;
  //add_record(&cache[r->clientno], *r); 
  if(r->prefetch == 0){
    fprintf(stdout, "t %f nettime %f\n", t, tmp);
    add_statistics(&st, t - r->start + 2*tmp);
  }
}

int io_request(struct disksim *disksim, Request io_request){
  completed = 0;
  now = io_request->start;
  //fprintf(stderr, "now is %f and the start is %lf\n", now, io_request->start );
  io_request->prefetch = 0; /*tempory */
  disksim_interface_request_arrive(disksim, now, io_request);
  /* Process events until this I/O is completed */
  fprintf(stderr, "next_event is %f\n", next_event);
  while(next_event >= 0) {
    now = next_event;
    next_event = -1;
    disksim_interface_internal_event(disksim, now);
    fprintf(stderr, "next_event is %f\n", next_event);
  }
  while(completed == 0){
    disksim_interface_internal_event(disksim, now);
  }
  
  if (!completed) {
    fprintf(stderr,
	    "%s: internal error. Last event not completed %d\n",
	    argv[0], i);
    exit(1);
  }
  return 0;
}

