#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <math.h>
#include "nasd.h"
#include "disksim_rand48.h"
#include "cache.h"
#include "network.h"
//#include "disksim_global.h"
#include "disksim_iotrace.h"

#define	BLOCK	4096
#define	SECTOR	512
#define	BLOCK2SECTOR	(BLOCK/SECTOR)


typedef	struct	{
	int n;
	double sum;
	double sqr;
} Stat;


static SysTime now = 0;		/* current time */
static SysTime next_event = -1;	/* next event */
static int completed = 0;	/* last request was completed */
static Stat st;

Cache client_cache[50];
Cache disk_cache[50];
FILE *fd1;
Trace_list *demand_q, *prefetch_q;
int client_hit_rate, disk_hit_rate;

void
panic(const char *s)
{
	perror(s);
	exit(1);
}


void
add_statistics(Stat *s, double x)
{
	s->n++;
	s->sum += x;
	s->sqr += x*x;
}


void
print_statistics(Stat *s, const char *title)
{
	double avg, std;

	avg = s->sum/s->n;
	std = sqrt((s->sqr - 2*avg*s->sum + s->n*avg*avg) / s->n);
	//fprintf(fd1, "%s: n=%d average=%f std. deviation=%f client hit = %d disk hit = %d\n", title, s->n, avg, std, client_hit_rate, disk_hit_rate);
}

/*parser. */
int parse_trace_file(FILE *fd, Request *req){
  ioreq_event *temp;
  
  temp = malloc(sizeof(ioreq_event));
  temp = iotrace_get_ioreq_event(fd, ASCII, temp);

  if(temp == NULL){
    return -1;
  };
  req->devno = temp->devno;
  req->blkno = temp->blkno;
  req->bcount = temp->bcount;
  req->flags = temp->flags;
  req->clientno = temp->clientno;
  req->start = temp->time;

  //fprintf(stdout, "check check. blkno is %d\n", req->blkno);
  free(temp);
  return 0;
  
}

int process_next_trace(ioreq_event * item){
  return 0;
}
/*
 * Schedule next callback at time t.
 * Note that there is only *one* outstanding callback at any given time.
 * The callback is for the earliest event.
 */
void
syssim_schedule_callback(void (*f)(void *,double), SysTime t)
{
	next_event = t;
}


/*
 * de-scehdule a callback.
 */
void
syssim_deschedule_callback(void (*f)())
{
	next_event = -1;
}


void
syssim_report_completion(SysTime t, Request *r)
{
  float tmp;
  double sim_time;
  Trace_list *temp_list, *next_list;
  
  tmp = net_time(r);
  completed = 1;
  now = t;

  r->wait = 0; /* for now trace only*/
  r->hit_count = 0;
  if(r->wait < 0) r->wait = 0;
  if(r->prefetch == 2){
    add_record(&disk_cache[r->devno], *r);
  }
  else if(r->prefetch == 0){ 
    /*find some space for the new record*/
    temp_list = client_cache[r->clientno].list;
    while(temp_list && client_cache[r->clientno].free_size < r->bcount){
      next_list = temp_list->next;
      //fprintf(stderr, "delete here\n");
      delete_record(&client_cache[r->clientno], temp_list->record);
      temp_list = next_list;
      //fprintf(stderr, "free size in cache is %d, and r needs %d\n",client_cache[r->clientno].free_size, r->bcount); 
    }
   
    // fprintf(stderr, "free size in cache is %d, and r needs %d\n",client_cache[r->clientno].free_size, r->bcount); 
    assert(client_cache[r->clientno].free_size >= r->bcount);
    add_record(&client_cache[r->clientno], *r);
    sim_time = now - r->start + r->wait + 2*tmp;
    //fprintf(fd1, "%f\t%f\t%f\n", r->start, sim_time, r->wait);
    add_statistics(&st, sim_time);
  }
  else if(r->prefetch == 1){
    add_record(&client_cache[r->clientno], *r); 
    /*fprintf(stdout, "request %d at device %d prefetched at time %f\n", r->blkno, r->devno, t);*/
  }
}


int disksim_io_request(struct disksim *disksim, Request *io_request){/* plan to put some priority later */

  /* if(io_request->prefetch == 1){
    push_queue(prefetch_q, io_request);
  }
  else{
    push_queue(demand_q, io_request);
  }

  if(demand_q){
    pop_queue(demand_q, io_request);
  }
  else{
    pop_queue(prefetch_q, io_request);
    }*/

  completed = 0;
  /*fprintf(stderr, "now is %f and the start is %lf, prefetch is %d, devno is %d\n", now, io_request->start, io_request->prefetch, io_request->devno );*/
  if(io_request->prefetch == 1){ /* little trick */
    io_request->start = now;
  }
  if((io_request->start-now) < 0.0001){
    io_request->start= now;
    io_request->wait += (now - io_request->start);
  }
  else{
    now = io_request->start;
  }
  //fprintf(stderr, "now is %f and the start is %lf\n", now, io_request->start );
  disksim_interface_request_arrive(disksim, now, io_request);
  /* Process events until this I/O is completed */
  //fprintf(stderr, "next_event is %f\n", next_event);
  while(next_event >= 0) {
    now = next_event;
    next_event = -1;
    disksim_interface_internal_event(disksim, now);
    // fprintf(stderr, "next_event is %f\n", next_event);
  }
  while(completed == 0){
    disksim_interface_internal_event(disksim, now);
  }
  
  if (!completed) {
    fprintf(stderr,
	    "internal error. Last event not completed\n");
    exit(1);
  }

  return 0;
}

int prefetch(Cache *cache, Request *record, int method){

   
  switch(method){
  case 0:
    return no_prefetch(cache, record); 
  case 1:
    return fixed_fetch(cache, record);
  case 2:
    return aggressive(cache, record); 
  case 3:
    return fixed_horizen(cache, record, st.n);
  }
  return 0;
}

void eat_white(char *string){
  char temp;
  int i;

  temp = string[0];
  for(i = 0; i< strlen(string); i++){
    if(temp == ' '){
      string = string +1;
    }
  }
}

int
main(int argc, char *argv[])
{
	int i, client_no,disk_no, ok, k, r, client_queueing, window;
	int client_cache_size, client_cache_method;
	int disk_cache_size, disk_cache_method;
	Request * io_request;
	long bandwidth;
	struct stat buf;
	struct disksim *disksim;
	int len = 8192000;
	FILE* fd, *fd2, *fd3;
	FILE* fp;
	char *disk_filename, *output_filename, *nasd_out_filename, *trace_filename;
	char *string1, *string2;
	float negative_error, positive_error, tmp, sim_time;
	char line[100];
	long state1[32] = {
	  3,
	  0x9a319039, 0x32d9c024, 0x9b663182, 0x5da1f342,
	  0x7449e56b, 0xbeb1dbb0, 0xab5c5918, 0x946554fd,
	  0x8c2e680f, 0xeb3d799f, 0xb11ee0b7, 0x2d436b86,
	  0xda672e2a, 0x1588ca88, 0xe369735d, 0x904f35f7,
	  0xd7158fd6, 0x6fa6f051, 0x616e6b96, 0xac94efdc,
	  0xde3b81e0, 0xdf0a6fb5, 0xf103bc02, 0x48f340fb,
	  0x36413f93, 0xc622c298, 0xf5a42ab8, 0x8a88d77b,
	  0xf5ad9d0e, 0x8999220b, 0x27fb47b9
	};

	setstate(state1); /* for random generator */
	
	if (argc != 2) {
	  fprintf(stderr, "usage: %s <diskparam file> \n", argv[0]);
	  exit(1);
	}

	if (stat(argv[1], &buf) < 0)
		panic(argv[1]);


	fp = fopen(argv[1], "r");
	while(fgets(line, 100, fp) != NULL) {
	  string1 = strtok(line, ":");
	  string2 = strtok(NULL, " \n");
	  //eat_white(string2);
	  if(strstr(string1, "disk parv")){
	    disk_filename = malloc(strlen(string2)+1);
	    strcpy(disk_filename, string2);
	    continue;
	  }
	  if(strstr(string1, "disksim output")){
	    output_filename = malloc(strlen(string2));
	    strncpy(output_filename, string2, strlen(string2));
	    continue;
	  }
	  if(strstr(string1, "nasd output")){
	    nasd_out_filename = malloc(strlen(string2));
	    strncpy(nasd_out_filename, string2, strlen(string2)); 
	    continue;
	  }
	  if(strstr(string1, "trace filename")){
	    trace_filename = malloc(strlen(string2));
	    strncpy(trace_filename, string2, strlen(string2)); 
	    continue;
	  }
	  if(strstr(string1, "clientno")){
	    client_no = atoi(string2);
	    continue;
	  }
	  if(strstr(string1, "diskno")){
	    disk_no = atoi(string2);
	    continue;
	  }
	  if(strstr(string1, "network bandwidth")){
	    bandwidth = atol(string2);
	    continue;
	  }
	  if(strstr(string1, "client cache size")){
	    client_cache_size = atoi(string2);
	    continue;
	  }
	  if(strstr(string1, "client cache method")){
	    client_cache_method = atoi(string2);
	    continue;
	  }
	  if(strstr(string1, "disk cache size")){
	    disk_cache_size = atoi(string2);
	    continue;
	  }
	  if(strstr(string1, "disk cache method")){
	    disk_cache_method = atoi(string2);
	    continue;
	  }
	  if(strstr(string1, "informed accuracy")){
	    negative_error = atof(string2);
	    continue;
	  }
	  if(strstr(string1, "positive informed")){
	    positive_error = atof(string2);
	    continue;
	  }
	  if(strstr(string1, "client queueing")){
	    client_queueing = atoi(string2);
	    continue;
	  }
	  if(strstr(string1, "sliding window")){
	    window = atoi(string2);
	    continue;
	  }
	}

        //fprintf(stderr, "client cache size is %d\n", client_cache_size);
        client_hit_rate = disk_hit_rate = 0;

	fd = fopen(trace_filename, "r");
	fd2 = fopen(trace_filename, "r");
	fd3 = fopen(trace_filename, "r");
	fd1 = fopen(nasd_out_filename, "w");
	assert(fd != NULL);
		
	init_network(bandwidth);
	disksim = malloc (len);
	disksim = disksim_interface_initialize(disksim, len, 
					       disk_filename, output_filename);
	for(i = 0; i< client_no; i++){
	  init_client_cache(&client_cache[i], i, client_cache_size, fd2, disksim);
	  client_cache[i].negative_error = 1.0 - negative_error;
	  client_cache[i].positive_error = positive_error;
	}
	for(i = 0; i< disk_no; i++){
	  init_disk_cache(&disk_cache[i], i, disk_cache_size, fd3, disksim);
	  disk_cache[i].negative_error =1.0 - negative_error;
	  disk_cache[i].positive_error = positive_error;
	}
	demand_q = prefetch_q = NULL;
      
	io_request = malloc(sizeof(Request));
	i = parse_trace_file(fd, io_request);
	for (; i != -1; i = parse_trace_file(fd, io_request) ) {
	  io_request->wait = 0;
	  /* do client side prefetch */
	  /*use spare time for prefetch*/
	  if(client_cache_method != 1){
	    while(now < (io_request->start - 0.005)){
	      ok = 0;
	      for(k = 0; k< client_no; k++){
		r = prefetch(&client_cache[k], io_request, client_cache_method);
		if(r == 1){
		  disk_cache[io_request->devno].last_prefetched = io_request->start;
		}
		ok = ok + r;
	      }
	      if(ok == 0) break; /*prefetch ends*/
	    }
	  }

	  /* do diskside prefetch */
	  // fprintf(stderr, "disk_cache_method is %d\n", disk_cache_method);
	  if(disk_cache_method != 1){
	    while(now < (io_request->start - 0.005)){
	      ok = 0;
	      for(k = 0; k< disk_no; k++){
		r = prefetch(&disk_cache[k], io_request, disk_cache_method);
		ok = ok + r;
	      }
	      
	      if(ok == 0) break; /*prefetch ends*/
	    }
	  }

	  ok = cache_lookup(&client_cache[io_request->clientno], *io_request);
	  if(!ok){ 
	    ok = cache_lookup(&disk_cache[io_request->devno], *io_request);
	    if(!ok){
                //fprintf(stderr, "can't find record in the cache\n");
	      if((io_request->start - now) < 0.0001){
                  io_request->wait = now - io_request->start;
                  io_request->start = now;
	      }
	      if(client_queueing == 3){
		io_request->wait = 0; /* for now trace*/
	      }
	      io_request->prefetch = 0;
	      if(client_cache_method == 1){
		io_request->bcount += 5;
	      }
	      disksim_io_request(disksim, io_request);
	    }
	    else{
                //fprintf(stderr, "Hit the disk cache\n");
              disk_hit_rate ++;
	      tmp = net_time(io_request);
	      now = io_request->start  + 2*tmp;// should plus network time
	      sim_time = now - io_request->start;
              io_request->prefetch = 0;
              if(disk_cache_method == 3){
                  update_hit_rate(&disk_cache[io_request->devno],*io_request);
              }
              else{
                  update_cache(&disk_cache[io_request->devno],*io_request);
              }
	      //fprintf(fd1, "%f\t%f\t0\n", io_request->start, sim_time);
	      add_statistics(&st, sim_time);
	    }
	  }
	  else{
              //fprintf(stderr, "Hit the local cache\n");
            client_hit_rate ++;
	    now = io_request->start;
            io_request->prefetch = 0;
            if(client_cache_method == 3){
                update_hit_rate(&client_cache[io_request->devno],*io_request);
            }
            else{

                update_cache(&client_cache[io_request->clientno],*io_request);
            }
	    //fprintf(fd1, "%f\t%f\t%f\n", io_request->start, 0,0);
	    add_statistics(&st, 0);
	  }

	}
	
	disksim_interface_shutdown(disksim, now);
	close_network();
	for(i = 0; i< client_no; i++){
	  close_client_cache(&client_cache[i]);
	}
	
	for(i = 0; i< disk_no; i++){
	  close_disk_cache(&disk_cache[i]);
	}

	print_statistics(&st, "response time");
	
	free(io_request);
	fclose(fd);
	fclose(fd1);
	fclose(fd2);
	fclose(fd3);
	exit(0);
}
